@@ -14,8 +14,30 @@ from util import get_files, register_keep_alive, run_function_list
 logger = logging.getLogger(__name__)
 
 
+def normalize_files_dict(files, path):
+    """
+    Sets up the paths for files dictionaries that will be ran through
+    get_files
+
+    get_files call with a dict must specify a path anyway.
+
+    :param files: the files dict/list of dicts to update
+    :param path: The path to add if it's missing
+    :return: void
+    """
+    # If files is a dict and no path is specified, set the path on files.
+    if isinstance(files, dict) and 'path' not in files:
+        files['path'] = path
+    # If files is a list of dicts and no path is specified on any of them, set the path for all
+    # missing path information
+    elif isinstance(files, list) and isinstance(files[0], dict):
+        for f in files:
+            if 'path' not in f:
+                f['path'] = path
+
+
 def watch(fns, path, files=None, run_immediately=True,
-          recursive=False, throttle=1, keep_alive=True, poll_timeout=0, handler=None):
+          recursive=False, throttle=1, keep_alive=True, poll_timeout=0, callback=None, handler=None):
     """
 
     Runs the given function list when items at path change
@@ -34,6 +56,7 @@ def watch(fns, path, files=None, run_immediately=True,
     :param poll_timeout: In certain setups, such as vagrant, FS notifications don't work.
                          Set this to a positive integer to force polling the FS.
                          This will also set the default blocking timeout.
+    :param callback: Callback to run after function list
     :param handler: the watch handler to use
 
     """
@@ -43,40 +66,27 @@ def watch(fns, path, files=None, run_immediately=True,
     if not path:
         raise Exception("Must specify a path to watch")
 
-    if not isinstance(fns, list):
-        fns = [fns, ]
-
     if run_immediately:
         run_function_list(fns)
 
-    # get_files call with a dict must specify a path anyway.
     if files:
-        # If files is a dict and no path is specified, set the path on files.
-        if isinstance(files, dict) and 'path' not in files:
-            files['path'] = path
-        # If files is a list of dicts and no path is specified on any of them, set the path for all
-        # missing path information
-        elif isinstance(files, list) and isinstance(files[0], dict):
-            for f in files:
-                if 'path' not in f:
-                    f['path'] = path
+        normalize_files_dict(files, path)
 
-    f = get_files(files) if files else None
-    handler = handler or WatchHandler(fns, 
-            throttle=throttle, 
-            predicate=lambda event: not f or event.src_path in f)
+    def predicate(event, watcher):
+        if watcher.files is None:
+            return True
+        files = get_files(watcher.files)
 
-    if poll_timeout:
-        observer = PollingObserver(timeout=poll_timeout)
-    else:
-        observer = Observer()
+        return event.src_path in files or getattr(event, 'dest_path', None) in files
 
-    observer.schedule(
-        handler,
-        path=path,
-        recursive=recursive
-    )
+    handler = handler or WatchHandler(fns,
+                                      throttle=throttle,
+                                      files=files,
+                                      callback=callback,
+                                      predicate=predicate)
 
+    observer = PollingObserver(timeout=poll_timeout) if poll_timeout else Observer()
+    observer.schedule(handler, path=path, recursive=recursive)
     observer.start()
 
     if keep_alive:
@@ -87,13 +97,14 @@ class WatchHandler(FileSystemEventHandler):
     """
     Default watch handler class
     """
-    def __init__(self, fns, throttle=1, callback=None, predicate=None):
+    def __init__(self, fns, files=None, throttle=1, callback=None, predicate=None):
         """
         initialize watch handler
         :param fns: function list to run on event
         :param callback: function to call
         """
         self.fns = fns
+        self.files = files
         self.callback = callback
         self.predicate = predicate
         self.throttle = throttle
@@ -107,12 +118,21 @@ class WatchHandler(FileSystemEventHandler):
         :type event:
             :class:`FileSystemEvent`
         """
-        logger.debug("Got {} watch dispatch for path {}. Event Type: {}".format(event.event_type,
-                                                                         event.src_path,
-                                                                         str(event.event_type)))
+        logger.debug("Got {} watch dispatch for path {}. Event Type: {}".format(
+                event.event_type,
+                event.src_path,
+                str(event.event_type)
+            )
+        )
 
-        if self.predicate and not self.predicate(event):
+        if self.last_run:
+            if (datetime.datetime.now() - self.last_run).total_seconds() <= self.throttle:
+                logger.debug("Throttle time not elapsed - not running event")
+                return
+
+        if self.predicate and not self.predicate(event, self):
             return
+
         super(WatchHandler, self).dispatch(event)
 
     def on_any_event(self, event):
@@ -122,14 +142,9 @@ class WatchHandler(FileSystemEventHandler):
         logger.info("Got {} event for path {}. Event Type: {}".format(event.event_type,
             event.src_path, str(event.event_type)))
 
-        if self.last_run:
-            if (datetime.datetime.now() - self.last_run).total_seconds() < self.throttle:
-                logger.info("Throttle time not elapsed - not running event")
-                return
-
-        self.last_run = datetime.datetime.now()
-
         try:
+            self.last_run = datetime.datetime.now()
+
             run_function_list(self.fns)
             if self.callback:
                 self.callback(event, self.fns)
